////////////////////////////////////////////////////////
// File:     deadlock.c 
// Purpose:  example deadlock situation
// Compile:  gcc -lpthread -Wall -g -o  deadlock deadlock.c
//
// Author:   Eric Kee
// Class:    CS78, Computer Networks, Spring 2008
//
// Comments:
//   This program demonstrates how deadlock can occur
//   when using locks.  In this scenario, Alice and Bob
//   acquire access to locks for the car and the keys.
//   I have inserted randomized sleeps to force different
//   execution orders to occur.  To see that deadlock
//   occurs, run this program a bunch of times--it will
//   eventually hang and you will have to hit ctrl-c to
//   kill it.
//
//   Deadlock occurs when there is a cycle of lock 
//   acquisitions.  In this case Alice tries to get the
//   keys while Bob already has the keys and is trying
//   to get the car.  In general, deadlock occurs 
//   whenever there is a cycle of un-fulfillable resource
//   requests (a lock can be seen as an abstract type of
//   resource).
//
////////////////////////////////////////////////////////
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>                                    // Including pthread API


// Globally shared variables
pthread_mutex_t car_lock = PTHREAD_MUTEX_INITIALIZER;   // Lock to control access the "car"
pthread_mutex_t key_lock = PTHREAD_MUTEX_INITIALIZER;   // Lock to control acces to the "keys"
int car = 0;                                            // The supposed "car"

int flip_coin(void);                                    // Borrow some functions so we can force
void consider_interleaving(void);                       //  ..random orderings to demonstrate problems

// None of these bad things happen, though, because we use locks
void* AliceFunc(void *ptr) 
{
  consider_interleaving();                              // Maybe we'll delay, maybe not...
  pthread_mutex_lock(&car_lock);                        // Get lock to get into the car

  consider_interleaving();                              // Maybe we'll delay here, maybe not...
  pthread_mutex_lock(&key_lock);                        // Get lock to pick up the keys

  car = 1000;                                           // Use the shared resource "car"
  printf("Alice:  Ok, i'm goin' %i miles an hour...\n", car);
  //sleep(1);                                           // Do stuff here
  printf("Alice:  Ok, i'm done with the car.\n");
 
  pthread_mutex_unlock(&car_lock);                      // Get lock to get into the car
  pthread_mutex_unlock(&key_lock);                      // Get lock to pick up the keys

  return NULL; 
}//end AliceFunc()


// This bad case won't ever happen, though...
void* BobFunc(void* ptr)
{
  consider_interleaving();                              // Maybe we'll delay here, maybe not...
  pthread_mutex_lock(&key_lock);                        // Get lock to pick up the keys

  consider_interleaving();                              // Maybe we'll delay here, maybe not...
  pthread_mutex_lock(&car_lock);                        // Get lock to get into the car

  car = 4;                                              // Use the shared resource "car"
  printf("Bob:  Ok, i'm changin' the %i tires... \n", car);
  //sleep(1);                                           // Do stuff here
  printf("Bob:  Ok, i'm done with the tires.\n");

  pthread_mutex_unlock(&key_lock);                      // Get lock to pick up the keys
  pthread_mutex_unlock(&car_lock);                      // Get lock to get into the car

  return NULL;
}//end BobFunc()


// Main will be our first thread
int main() 
{
  int rc;                                               // For return codes
  pthread_t AliceThread;                                // Two thread variables for
  pthread_t BobThread;                                  //  ..Alice and Bob
   
  // Create alice and bob
  rc = pthread_create(&AliceThread, NULL, AliceFunc, NULL);
  rc = pthread_create(&BobThread, NULL, BobFunc, NULL);

  // Wait for Alice and Bob to complete
  rc = pthread_join(AliceThread, NULL);       
  rc = pthread_join(BobThread, NULL);

  exit(0);                                              // Never reached
}//end main










//--------------------------------------------------------
// Source: Sean Smith (thanks!)
// first call: initialize random_byte
// next 8 calls: return a random bit
//---------------------------------------------------
int flip_coin(void) 
{

  static int  fd;
  static int initialized = 0;
  static int bits_left = 0;
  static unsigned char random_byte;
  int random_bit;

  if (!initialized) {
    
    fd = open("/dev/random", O_RDONLY);
    if (fd < 0) {
      fprintf(stderr,"open failed\n");
      exit(-1);
    }

    initialized = 1;
  }

  if (0 == bits_left) {
    if (sizeof(random_byte) != read(fd, &random_byte, sizeof(random_byte))) {
      fprintf(stderr, "read failed\n");
      exit(-1);
    }
    bits_left = 8;
  }

  random_bit = random_byte & 0x01;
  random_byte >>= 1;
  bits_left--;

  return random_bit;
}//end flip_coin(...)


// Source: Sean Smith
// Call this function to enter a randomized sleep
void consider_interleaving(void) 
{
  if (flip_coin())
    sleep(1);
}//end consider_interleaving(...)
